/**
 * Titanium SDK Library for Node.js
 * Copyright (c) 2012-2013 by Appcelerator, Inc. All Rights Reserved.
 * Please see the LICENSE file for information about licensing.
 */

var appc = require('node-appc'),
	i18n = appc.i18n(__dirname),
	__ = i18n.__,
	afs = appc.fs,
	fs = require('fs'),
	path = require('path');

function stripQuotes(s) {
	return (''+s).replace(/^(['"])(.+)(['"])$/, '$2');
}

exports.load = function (dir, deviceFamilies, logger, callback) {
	var entries = {};

	afs.visitFiles(dir, function (file, full) {
		if (/.+\.jss$/.test(file)) {
			var base = file.substring(0, file.length - 4).split('.'),
				density = base.pop(),
				platform = base.pop();
				entry = {
					base: null,
					platform: null,
					density: {}
				};

			if (deviceFamilies.indexOf(platform) == -1) {
				platform && base.push(platform);
				// maybe the density is our platform?
				if (deviceFamilies.indexOf(density)) {
					density && base.push(density);
					platform = null;
				} else {
					platform = density;
				}
				density = null;
			}

			base = base.join('.');
			entries[base] && (entry = entries[base]);

			if (density) {
				entry.density[density] = full;
			} else if (platform) {
				entry.platform = full;
			} else {
				entry.base = full;
			}

			entries[base] = entry;
		}
	}, function () {
		var results = {},
			parse = require('css-parse'),
			propMap = {
				'background-image': 'backgroundImage',
				'background-url': 'backgroundImage',
				'text-align': 'textAlign',
				'border-radius': 'borderRadius',
				'border-color': 'borderColor',
				'border-width': 'borderWidth',
				'background-color': 'backgroundColor'
			},
			parseFile = function (file, name, density) {
				logger.debug(__('Parsing JSS file: %s', file.cyan));
				var rules = parse(fs.readFileSync(file).toString()).stylesheet.rules;
				rules.forEach(function (rule) {
					if (rule['import']) {
						var importFilename = stripQuotes(rule['import']),
							importFile = path.join(path.dirname(file), importFilename);
						if (!afs.exists(importFile)) {
							logger.error(__('Error parsing JSS file.') + '\n');
							logger.log(__("Couldn't find import file %s referenced from %s", importFilename.cyan, file.cyan) + '\n');
							process.exit(1);
						} else {
							parseFile(importFile, name, density);
						}
					} else if (rule.selectors) {
						rule.selectors.forEach(function (selector) {
							var c = selector.charAt(0),
								dest = c == '#' ? 'ids' : c == '.' ? 'classes' : 'tags';

							dest != 'tags' && (selector = selector.substring(1));
							density && (dest += '_density');

							rule.declarations.forEach(function (decl) {
								var prop = propMap[decl.property] || decl.property,
									tmp;

								results.hasOwnProperty(dest) || (results[dest] = {});
								results[dest].hasOwnProperty(name) || (results[dest][name] = {});

								if (density) {
									results[dest][name].hasOwnProperty('density') || (results[dest][name][density] = {});
									results[dest][name][density].hasOwnProperty(selector) || (results[dest][name][density][selector] = {});
									tmp = results[dest][name][density][selector];
								} else {
									results[dest][name].hasOwnProperty(selector) || (results[dest][name][selector] = {});
									tmp = results[dest][name][selector];
								}

								if (/^font-.+/.test(prop)) {
									// font-size, font-family, font-weight, and friends are special and need to go in special font object
									tmp = tmp.font || (tmp.font = {});
									prop = 'font' + prop.split('-').map(function (p) { return p.substring(0, 1).toUpperCase() + p.substring(1); }).join('').substring(4);
								}

								tmp[prop] = stripQuotes(decl.value)
							});
						});
					}
				});
			};

		if (Object.keys(entries).length) {
			logger.info(__('Loading and parsing JSS files'));

			Object.keys(entries).forEach(function (name) {
				var entry = entries[name];
				entry.base && parseFile(entry.base, name);
				entry.platform && parseFile(entry.platform, name);
				Object.keys(entry.density).forEach(function (d) {
					parseFile(entry.density[d], name, d);
				});
			});
		}

		callback(results);
	});
};
